import os


def validate_process(pid, name=None):
    """
    Check whether there is a process with `pid` and `name`

    :param pid: pid of the running process
    :param name: name of the running process
    :type name: str or None
    :returns: whether there is a process with the given name and pid
    :rtype: bool
    """
    try:
        with open(os.path.join('/proc', str(pid), 'comm')) as proc_comm:
            proc_name = next(proc_comm).strip()
    except (OSError, IOError):
        return False
    return name is None or name == proc_name



from __future__ import absolute_import
import logging
import glob
import platform

import include

from .. import conf


_LOGGER = logging.getLogger(__name__)


def run_configurations(config_globs, pipelines=None):
    _LOGGER.warning('running configuration')
    _prev_pipelines = conf.pipelines
    conf.pipelines = pipelines = pipelines if pipelines is not None else []
    _LOGGER.info('%-16s => %s %s', 'interpreter', platform.python_implementation(), platform.python_version())
    _LOGGER.info('%-16s => %s', 'globs', config_globs)
    _LOGGER.info('%-16s => <%s> %s', 'pipelines', pipelines.__class__.__name__, pipelines)
    for glob_path in config_globs:
        for config_path in glob.iglob(glob_path):
            _LOGGER.info('configuration: %s', config_path)
            include.path(config_path)
            if conf.pipelines is not pipelines:
                pipelines = conf.pipelines
                _LOGGER.info('%-16s => <%s> %s', 'pipelines', pipelines.__class__.__name__, pipelines)
    conf.pipelines = _prev_pipelines
    return pipelines



from __future__ import absolute_import
import os
import logging
import argparse

from . import __about__
from .conf import loader, logger
from . import driver

_LOGGER = logging.getLogger(__name__)


def env_key(cli_key):
    _key = cli_key.strip('-')
    _key = __about__.__title__ + '-' + _key
    _key = _key.replace('-', '_')
    return _key.upper()


CLI = argparse.ArgumentParser(__about__.__title__)
CLI_CONFIG = CLI.add_argument_group('configuration options')
CLI_CONFIG.add_argument(
    'configuration',
    metavar='CONFIGURATION',
    nargs='*',
    default=['/etc/%s/*.py' % __about__.__title__],
    help='configuration file paths or globs [%(default)s]',
)
CLI_LOGGING = CLI.add_argument_group('logging options')
CLI_LOGGING.add_argument(
    '-l', '--log-level',
    metavar='LEVEL',
    default=os.environ.get(env_key('log-level'), 'WARNING'),
    help='logging verbosity, numeric or name [%%(default)s] ($%s)' % env_key('log-level'),
)
CLI_LOGGING.add_argument(
    '-f', '--log-format',
    metavar='FORMAT',
    help='logging message format [%%(default)s] ($%s)' % env_key('log-format'),
    default=os.environ.get(env_key('log-format'), '%(asctime)s (%(process)d) %(levelname)8s: %(message)s'),
)
CLI_LOGGING.add_argument(
    '-d', '--log-dest',
    metavar='DEST',
    help='logging destinations, as file path or stderr/stdout [%%(default)s] ($%s)' % env_key('log-dest'),
    nargs='*',
    default=[elem.strip() for elem in os.environ.get(env_key('log-dest'), 'stderr').split(',')]
)

options = CLI.parse_args()

_LOGGER.warning('## %s [v%s] %s' % (
    __about__.__title__, __about__.__version__, __about__.__url__)
)
logger.configure_logging(log_level=options.log_level, log_format=options.log_format, log_dest=options.log_dest)
for opt_name in ('configuration', 'log_level', 'log_dest', 'log_format'):
    _LOGGER.info('%-16s => %r', opt_name, getattr(options, opt_name))
pipelines = loader.run_configurations(options.configuration)
pipeline_driver = driver.PipelineDriver()
for pipeline in pipelines:
    pipeline_driver.mount(pipeline)
pipeline_driver.run()



from __future__ import absolute_import
import os
import errno
import time

import chainlet


@chainlet.genlet(prime=False)
def readlines(filelike):
    """
    Read lines from a file-like object

    Stops once the underlying object no longer provides any lines.
    """
    for line in filelike:
        yield line[:-1]


@chainlet.genlet(prime=False)
def tail_path(path, follow=True, **open_kwargs):
    """
    Stream data from a file, socket or similar object at ``path``

    Reads and provides lines of data from ``path`` (excluding line breaks).
    This works similar to the ``tail -F`` unix utility.

    :param path: the file system location suitable for :py:func:`open`
    :type path: str
    :param follow: whether to proceed reading from a new file if the original one is replaced
    :type follow: bool
    :param open_kwargs: keyword arguments to pass to :py:func:`open`

    Note that ``follow`` does not require immediate replacement with a *new* file.
    In ``follow`` mode, any errors from ``path`` not existing are ignored.
    Without ``follow`` mode, it is an error if ``path`` does not exist when the first chunk is fetched.
    Any other errors from :py:func:`open` are propagated unconditionally.
    """
    if follow:
        eof_delay = 0.01
        while True:
            try:
                for chunk in _tail_path_once(path, **open_kwargs):
                    yield chunk
            except OSError as err:
                if err.errno != errno.ENOENT:
                    raise
                time.sleep(eof_delay)
                eof_delay = min(0.5, eof_delay * 2)
            else:
                # if the file has been properly closed, reset delay
                eof_delay = 0.01
    else:
        for chunk in _tail_path_once(path, **open_kwargs):
            yield chunk


def _tail_path_once(path, **open_kwargs):
    eof_delay = 0.01
    path_inode = os.stat(path).st_ino
    with open(path, **open_kwargs) as stream:
        while True:
            line = stream.readline()
            if not line:
                # File is at EOF
                # - the target has been deleted or rotated
                if not os.path.exists(path) or os.stat(path).st_ino != path_inode:
                    break
                # - the file is still valid, smoothly delay further reads
                else:
                    time.sleep(eof_delay)
                    eof_delay = min(0.5, eof_delay * 2)
            else:
                eof_delay = 0.01
                yield line[:-1]  # strip away linebreak



import weakref
import threading


class Singleton(object):
    """
    Basic implementation of a Singleton

    Any instances constructed with the same parameters are actually the same object.
    Use :py:meth:`~.__singleton_signature__` to specify which parameters define identity.
    """
    __singleton_store__ = weakref.WeakValueDictionary()
    __singleton_mutex__ = threading.RLock()

    @classmethod
    def __singleton_signature__(cls, *args, **kwargs):
        # args is always a tuple, but kwargs is mutable and arbitrarily sorted
        return cls, args, tuple(sorted(kwargs.items()))

    def __new__(cls, *args, **kwargs):
        identifier = cls.__singleton_signature__(*args, **kwargs)
        with cls.__singleton_mutex__:
            try:
                self = cls.__singleton_store__[identifier]
            except KeyError:
                self = object.__new__(cls)
                self_init = self.__init__
                self.__singleton_init__ = True

                def singleton_init(*iargs, **ikwargs):
                    """Wrapper to run init only once for each singleton instance"""
                    if self.__singleton_init__:
                        self_init(*iargs, **ikwargs)
                        self.__singleton_init__ = False
                self.__init__ = singleton_init
                self.__init__.__doc__ = self_init.__doc__
                cls.__singleton_store__[identifier] = self
            return self



from __future__ import division, absolute_import
import logging

import chainlet.driver


class PipelineDriver(chainlet.driver.ThreadedChainDriver):
    """
    Driver for processing pipelines
    """
    def __init__(self):
        super(PipelineDriver, self).__init__()
        self._logger = logging.getLogger('%s.%s' % (__name__, self.__class__.__name__))

    def run(self):
        """
        Collect and pass on reports
        """
        self._logger.info('driving %d pipeline(s)', len(self.mounts))
        self._logger.info('starting %s main loop', self.__class__.__name__)
        super(PipelineDriver, self).run()
        self._logger.info('stopping %s main loop', self.__class__.__name__)



import time

import chainlet


@chainlet.funclet
def delay(value, duration):
    """
    Pause a chain for ``duration`` seconds

    :param value: arbitrary data chunk
    :param duration: seconds to pause chain execution
    :type duration: int or float
    :return:
    """
    time.sleep(duration)
    return value


class Every(chainlet.ChainLink):
    """
    Pass on only every nth value

    :param nth: the repeating index of the data chunk to pass on
    :param first: how many chunks to wait before passing on the first
    """
    def __init__(self, nth=10, first=0):
        self.nth = nth
        self._count = 0 - first

    def chainlet_send(self, value=None):
        count = self._count = self._count + 1
        if count % self.nth == 1 and count > 0:
            return value
        raise chainlet.StopTraversal



"""
Tools to interact with XRootD's `report` functionality
"""
from __future__ import division, absolute_import
import time
import logging
import subprocess
import atexit

from ..utilities import singleton, safe_eval

import chainlet


class XRootDReports(singleton.Singleton, chainlet.ChainLink):
    """
    Collects information generated by the `all.report` directive

    :param port: port receiving reports
    :type port: int

    Provides xrootd reports as individual dictionaries to a chain.
    Keys are preserved, while values are evaluated as literals.
    """
    def __init__(self, port):
        super(XRootDReports, self).__init__()
        self.port = port
        self._reportstreamer = None
        self._logger = logging.getLogger('%s.%s' % (__name__, self.__class__.__name__))

    @classmethod
    def __singleton_signature__(cls, port):
        return XRootDReports, port

    def open(self):
        """Start collecting reports"""
        if self._reportstreamer is None:
            self._logger.info('opening report stream on port %d', self.port)
            self._reportstreamer = subprocess.Popen(
                ['mpxstats', '-p', str(self.port), '-f', 'cgi'],
                stdout=subprocess.PIPE,
                universal_newlines=True,
            )
            self._logger.info('buffering report stream')
            atexit.register(self._reportstreamer.terminate)

    def close(self):
        """Stop collecting reports"""
        if self._reportstreamer is not None:
            if self._reportstreamer.poll() is None:
                self._reportstreamer.terminate()
                time.sleep(0.1)
            self._logger.info('closed report stream on port %d (exit code: %s)', self.port, self._reportstreamer.poll())
            self._reportstreamer = None

    def chainlet_send(self, value=None):
        """Fetch a report"""
        self.open()
        line = self._reportstreamer.stdout.readline()
        self._logger.debug('received datagram: %r', line)
        datagram = dict(item.split('=') for item in line.split('&'))
        for key in datagram:
            datagram[key] = safe_eval(datagram[key])
        return datagram

    def __enter__(self):
        # start collection on entering the context, do not delay until consumption
        self.open()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()
        return False

    def __del__(self):
        self.close()


xrdreports = XRootDReports



__title__ = 'pypelined'
__summary__ = 'stream and pipeline processing service'
__url__ = 'https://github.com/maxfischer2781/pypelined'

__version__ = '0.1.3'
__author__ = 'Max Fischer'
__email__ = 'maxfischer2781@gmail.com'
__copyright__ = '2017 %s' % __author__



"""
Backend for sending reports using ApMon for the ALICE collaboration
"""
from __future__ import division, absolute_import

import logging
import socket
import time

import apmon
import chainlet
import chainlet.dataflow

from ..utilities import proctools
from ..utilities import dfs_counter


class ApMonLogger(object):
    """
    Replacement for ApMon `Logger.Logger` class to use default python logging

    :param defaultLevel: ignored, available for signature compatibility only

    Redirects ApMon logging calls messages using the standard :py:mod:`logging`
    utilities. This is intended for compatibility with backend code, mainly
    ApMon itself. If you wish to change ApMon logging settings, directly modify
    the logger `"xrdmonlib.ApMonLogger"`.
    """
    #: log levels as used by ApMon Logger
    apmon_levels = ('FATAL', 'ERROR', 'WARNING', 'INFO', 'NOTICE', 'DEBUG')
    #: map from apmon levels to logging levels
    level_map = {
        0: logging.CRITICAL,
        1: logging.ERROR,
        2: logging.WARNING,
        3: logging.INFO,
        4: logging.DEBUG,
        5: logging.DEBUG,
    }

    def __init__(self, defaultLevel=None):
        self._logger = logging.getLogger('%s.%s' % (__name__.split('.')[0], self.__class__.__name__))
        self._logger.warning('redirecting ApMon logging to native logger %r', self._logger.name)

    def log(self, level, message, printex=False):
        """Log a message from ApMon"""
        self._logger.log(self.level_map[level], '[ApMon] %s', message, exc_info=printex)

    def setLogLevel(self, strLevel):
        """Set the logging level"""
        logging_level = self.level_map[self.apmon_levels.index(strLevel)]
        self._logger.setLevel(logging_level)
        self._logger.log(logging_level, 'logging level set via ApMon: %s => %s', strLevel, logging_level)
for _level, _name in enumerate(ApMonLogger.apmon_levels):
    setattr(ApMonLogger, _name, _level)


class ApMonReport(dict):
    """
    Container for named reports for ApMon following ALICE conventions

    :param cluster_name: identifier for the host group this report applies to
    :type cluster_name: str
    :param node_name: identifier for the host this report applies to
    :type node_name: str
    :param params: parameters to send
    :type params: dict
    """
    def __init__(self, cluster_name, node_name, params):
        super(ApMonReport, self).__init__(self)
        self.cluster_name = cluster_name
        self.node_name = node_name
        self.update(params)


class Reporter(chainlet.ChainLink):
    """
    BaseClass for converters from report dicts to :py:class:`ApMonReport`
    """
    def __init__(self):
        super(Reporter, self).__init__()
        self._logger = logging.getLogger('%s.%s' % (__name__, self.__class__.__name__))
        self._hostname = socket.getfqdn()

    def _xrootd_cluster_name(self, report):
        """Format report information to create ALICE cluster name"""
        return '%(se_name)s_%(instance)s_%(daemon)s_Services' % {
            'se_name': report['site'],  # must be defined via all.sitename
            'instance': report['ins'],
            'daemon': report['pgm'],
        }


class XrootdSpaceReporter(Reporter):
    """
    Extract XRootD Space information from reports

    :param weight_reporters: weight space statistics by number of reporter hosts
    :type weight_reporters: bool

    Provides key statistics on available space on an xrootd oss:

    *xrootd_version*
        The version of the xrootd daemon

    *space_total*
        The total space in MiB

    *space_free*
        The available space in MiB

    *space_largestfreechunk*
        The largest, consecutive available space in MiB
    """
    def __init__(self, weight_reporters=True):
        super(XrootdSpaceReporter, self).__init__()
        self._space_counters = {}
        self.weight_reporters = weight_reporters

    def chainlet_send(self, value=None):
        if 'oss.paths' not in value:
            raise chainlet.StopTraversal
        xrootd_report = self._oss_path_stats(report=value)
        return ApMonReport(cluster_name=self._xrootd_cluster_name(value), node_name=self._hostname, params=xrootd_report)

    def _oss_path_stats(self, report):
        """Collect statistics from ``oss.path`` fields"""
        try:
            path_count = report['oss.paths']
        except KeyError:
            raise chainlet.StopTraversal
        path_stats = {
            'xrootd_version' : report['ver'],
            'space_total': 0,
            'space_free': 0,
            'space_largestfreechunk': 0,
        }
        for path_id in range(path_count):
            # get real path to volume to count how many reporters see it
            path_rp = report['oss.paths.%d.rp' % path_id]
            path_reporters = self._get_path_share(path_rp)
            self._logger.debug('adding report (%d reporters) for path %r', path_reporters, path_rp)
            # reports are in kiB, MonALISA expects MiB
            path_stats['space_total'] += report['oss.paths.%d.tot' % path_id] / path_reporters / 1024
            path_stats['space_free'] += report['oss.paths.%d.free' % path_id] / path_reporters / 1024
            path_stats['space_largestfreechunk'] = max(
                path_stats['space_largestfreechunk'],
                report['oss.paths.%d.free' % path_id] / 1024
            )
        return path_stats

    def _get_path_share(self, path):
        """Get the number of hosts reporting the same space share"""
        if path not in self._space_counters:
            self._space_counters[path] = dfs_counter.DFSCounter(path)
        return self._space_counters[path]


class AliceApMonBackend(Reporter):
    """
    Backend for ApMon client to MonALISA Monitoring

    :param destination: where to send data to, as `"hostname:port"`
    :type destination: str
    """
    def __init__(self, *destination):
        super(AliceApMonBackend, self).__init__()
        self._logger = logging.getLogger('%s.%s' % (__name__, self.__class__.__name__))
        # initialization
        self.destination = destination
        # initialize ApMon, reroute logging by replacing Logger at module level
        apmon_logger, apmon.Logger = apmon.Logger, ApMonLogger
        self._apmon = apmon.ApMon(destination)
        apmon.Logger = apmon_logger
        # BUGFIX: apmon can create an invalid identifier on systems with pids > 32767
        if any(senderRef['INSTANCE_ID'] > 2147483647 for senderRef in self._apmon.senderRef.values()):
            raise RuntimeError('invalid ApMon INSTANCE_ID')  # https://github.com/MonALISA-CIT/apmon_py/issues/4
        # background monitoring
        self._background_monitor_sitename = None
        self._service_job_monitor = set()

    def chainlet_send(self, value=None):
        """Send reports via ApMon"""
        # preformatted report
        if self._send_apmon_report(value):
            return value
        # regular xrootd report
        self._send_raw_report(value)
        # configure background monitoring of xrootd components from report
        self._monitor_host(value)
        self._monitor_service(value)

    def _send_apmon_report(self, value):
        """Send a report preprocessed for apmon"""
        try:
            cluster_name = value.cluster_name
            node_name = value.node_name
        except AttributeError:
            return False
        else:
            self._apmon.sendParameters(
                clusterName=cluster_name,
                nodeName=node_name,
                params=value
            )
            self._logger.info('apmon report for %r @ %r sent to %s' % (cluster_name, node_name, str(self.destination)))
            return True

    def _send_raw_report(self, value):
        """Send a raw report from xrootd"""
        self._send_apmon_report(ApMonReport(
            cluster_name='%(se_name)s_xrootd_ApMon_Info' % {'se_name': value['site']},
            node_name=self._hostname,
            params=value
        ))

    def _monitor_host(self, value):
        """Add background monitoring for this host"""
        try:
            se_name = value['site']
        except KeyError:
            return False
        if self._background_monitor_sitename == se_name:
            return
        # configure background monitoring
        cluster_name = '%(se_name)s_xrootd_SysInfo' % {'se_name': se_name}
        self._apmon.setMonitorClusterNode(
            cluster_name,
            self._hostname
        )
        self._apmon.enableBgMonitoring(True)
        self._background_monitor_sitename = se_name
        self._logger.info(
            'apmon host monitor for %r @ %r added to %s' % (cluster_name, self._hostname, str(self.destination))
        )
        return True

    def _monitor_service(self, report):
        """Add background monitoring for a service"""
        if 'pgm' not in report or 'pid' not in report:
            return
        pid, now = int(report['pid']), time.time()
        # add new services for monitoring
        if pid not in self._service_job_monitor:
            cluster_name = self._xrootd_cluster_name(report)
            self._apmon.addJobToMonitor(
                pid=pid,
                workDir='',
                clusterName=cluster_name,
                nodeName=self._hostname,
            )
            self._service_job_monitor.add(pid)
            self._logger.info(
                'apmon job monitor for %r @ %r added to %s' % (cluster_name, self._hostname, str(self.destination))
            )
        for pid in list(self._service_job_monitor):
            if not proctools.validate_process(pid):
                self._apmon.removeJobToMonitor(pid)
                self._service_job_monitor.discard(pid)


# full ALICE monitoring backend stack
def alice_xrootd(*destinations):
    """
    Factory for ALICE XRootD ApMon/MonALISA Backend

    :param destinations: where to send data to, as `"hostname:port"`
    :type destinations: str
    """
    backend = AliceApMonBackend(*destinations)
    return (XrootdSpaceReporter(), chainlet.dataflow.NoOp()) >> backend



import ast

__all__ = ['safe_eval']


def safe_eval(literal):
    """
    Evaluate a literal value or fall back to string

    Safely performs the evaluation of a literal. If the literal is not valid,
    it is assumed to be a regular string and returned unchanged.

    :param literal: literal to evaluate, e.g. `"1.0"` or `"{'foo': 3}"`
    :type literal: str
    :return: evaluated or original literal
    """
    try:
        return ast.literal_eval(literal)
    except (ValueError, SyntaxError):
        return literal



import socket
import os
import sys
import hashlib
import filelock
import glob
import time
import threading
import logging
import weakref
import operator
import random
import errno

from .singleton import Singleton


#: Identifier for the process owning a counter
OWNER_IDENTIFIER = '%s\t%s\t%s' % (socket.getfqdn(), os.getpid(), sys.executable)


# actually a method of DFSCounter
# must be separate to allow garbage collection of self
def _count_updater(self):
    """separate counter loop to regularly check/verify state"""
    assert isinstance(self, weakref.ProxyTypes), "counter thread must receive weakref'd self to be collectible"
    # locally rebind everything we need to work
    self_repr = self.__repr__().replace('value=None', 'value=?')
    marker_path = self._marker_path
    thread_shutdown = self._thread_shutdown
    host_lock = self._host_lock
    logger = self._logger
    self._logger.info('acquiring %r @ %r', self_repr, marker_path)
    with host_lock:
        while not thread_shutdown.is_set():
            try:
                with open(marker_path, 'w') as marker:
                    marker.write(OWNER_IDENTIFIER)
                    logger.debug('marking %r @ %r', self_repr, marker_path)
                self._count_value = self._get_count()
            except ReferenceError:
                pass
            except Exception as err:
                logger.warning('failed updating %r: %s', self_repr, err)
            # wait for a fraction of timeout to allow write failures
            # jitter wait to smooth out path access
            thread_shutdown.wait(self.timeout / (3 + random.random()))
        logger.info('releasing %r @ %r', self_repr, marker_path)
        if os.path.exists(marker_path) and os.path.isfile(marker_path):
            os.unlink(marker_path)


class DFSCounter(Singleton):
    """
    Counter for hosts accessing the same Distributed File System

    :param shared_path: path used for synchronisation
    :type shared_path: str
    :param timeout: maximum age of synchronisation in seconds before assuming stale processes
    :type timeout: int or float

    This class pretends to be an integer for all operators. It counts the number
    of processes accessing the `shared_path`, and directly represents the result.
    Only one process per host may claim the `shared_path` at any time.
    """
    def __init__(self, shared_path, timeout=300):
        self._logger = logging.getLogger('%s.%s' % (__name__, self.__class__.__name__))
        self._thread_shutdown = threading.Event()
        self._thread_shutdown.clear()
        if not os.path.exists(os.path.dirname(shared_path)):
            raise OSError(errno.ENOENT, 'no such directory: %s' % os.path.dirname(shared_path))
        self.shared_path = shared_path
        self.timeout = timeout
        self._host_identifier = socket.getfqdn()
        self._marker_path = self._get_marker_path(hashlib.sha1(self._host_identifier.encode()).hexdigest())
        self._host_lock = filelock.FileLock(self._marker_path + '.lock')
        self._count_value = None
        self._thread = threading.Thread(target=_count_updater, args=(weakref.proxy(self),))
        self._thread.start()
        self._acquire()

    @classmethod
    def __singleton_signature__(cls, shared_path, timeout=300):
        return DFSCounter, shared_path

    def _acquire(self):
        # block until we own the resource
        block_delay = 0.005
        # we need to both OWN the resource (lock) and GET it as well (counter)
        while not self._host_lock.is_locked or self._count_value is None:
            self._logger.debug('waiting for exclusive host lock @ %r', self._marker_path)
            time.sleep(block_delay)
            block_delay = min(block_delay * 2, 5)
        self._logger.debug('acquired %r', self)

    def _get_marker_path(self, identifier):
        return '%s.dfsc-%s.csv' % (self.shared_path, identifier)

    # cross host counting
    def _get_count(self):
        min_age = time.time() - self.timeout
        return sum(1 for file_path in glob.iglob(self._get_marker_path('*')) if os.stat(file_path).st_mtime > min_age)

    def release(self):
        """Release the underlying resource"""
        self._thread_shutdown.set()

    def __del__(self):
        self._thread_shutdown.set()

    # representation
    def __str__(self):
        return str(int(self))

    def __repr__(self):
        return '<%s(shared_path=%r, timeout=%s), value=%s>' % (
            type(self).__name__, self.shared_path, self.timeout, self._count_value
        )

    # overriding __class__ at the instance level does not work, but changing the slot wrapper does
    @property
    def __class__(self):
        return int

    # comparisons
    def __lt__(self, other):
        return int(self) < other

    def __le__(self, other):
        return int(self) <= other

    def __eq__(self, other):
        return int(self) == other

    def __ne__(self, other):
        return int(self) != other

    def __gt__(self, other):
        return int(self) > other

    def __ge__(self, other):
        return int(self) >= other

    # value may change, not hashable
    __hash__ = None

    # number interface
    def __int__(self):
        return self._count_value

    __index__ = __int__

    def __complex__(self):
        return complex(int(self))

    def __float__(self):
        return float(int(self))

    def __long__(self):
        return long(int(self))

    def __round__(self, ndigits):
        return round(int(self), ndigits=ndigits)

    def __bytes__(self):
        return bytes(int(self))

    def __oct__(self):
        return oct(int(self))

    def __hex__(self):
        return hex(int(self))

    def __neg__(self):
        return -int(self)

    def __pos__(self):
        return int(self)

    def __abs__(self):
        return abs(int(self))

    def __invert__(self):
        return ~int(self)

    def __add__(self, other):
        return int(self) + other

    def __radd__(self, other):
        return other + int(self)

    def __sub__(self, other):
        return int(self) - other

    def __rsub__(self, other):
        return other - int(self)

    def __mul__(self, other):
        return int(self) * other

    def __rmul__(self, other):
        return other * int(self)

    def __floordiv__(self, other):
        return int(self) // other

    def __rfloordiv__(self, other):
        return other // int(self)

    def __mod__(self, other):
        return int(self) % other

    def __rmod__(self, other):
        return other % int(self)

    def __divmod__(self, other):
        return divmod(int(self), other)

    def __rdivmod__(self, other):
        return divmod(other, int(self))

    def __pow__(self, other, modulo):
        return pow(int(self), other, modulo)

    def __rpow__(self, other):
        return other ** int(self)

    def __lshift__(self, other):
        return int(self) << other

    def __rlshift__(self, other):
        return other << int(self)

    def __rshift__(self, other):
        return int(self) >> other

    def __rrshift__(self, other):
        return other >> int(self)

    def __and__(self, other):
        return int(self) & other

    def __rand__(self, other):
        return other & int(self)

    def __xor__(self, other):
        return int(self) ^ other

    def __rxor__(self, other):
        return other ^ int(self)

    def __or__(self, other):
        return int(self) | other

    def __ror__(self, other):
        return other | int(self)

    def __div__(self, other):
        return operator.div(int(self), other)

    def __rdiv__(self, other):
        return operator.div(other, int(self))

    def __truediv__(self, other):
        return operator.truediv(int(self), other)

    def __rtruediv__(self, other):
        return operator.truediv(other, int(self))



#!/usr/bin/env python
import os
from setuptools import setup, find_packages

repo_base_dir = os.path.abspath(os.path.dirname(__file__))
# pull in the packages metadata
package_about = {}
with open(os.path.join(repo_base_dir, "pypelined", "__about__.py")) as about_file:
    exec(about_file.read(), package_about)


with open(os.path.join(repo_base_dir, 'README.rst'), 'r') as README:
    long_description = README.read()

if __name__ == '__main__':
    setup(
        name=package_about['__title__'],
        version=package_about['__version__'],
        description=package_about['__summary__'],
        long_description=long_description.strip(),
        author=package_about['__author__'],
        author_email=package_about['__email__'],
        url=package_about['__url__'],
        packages=find_packages(),
        # dependencies
        install_requires=[
            'apmon',
            'filelock',
            'chainlet>=1.2.0',
            'include'
        ],
        # metadata for package search
        license='MIT',
        # https://pypi.python.org/pypi?%3Aaction=list_classifiers
        classifiers=[
            'Development Status :: 4 - Beta',
            'Intended Audience :: Developers',
            'Intended Audience :: System Administrators',
            'Topic :: System :: Monitoring',
            'Programming Language :: Python :: 2',
            'Programming Language :: Python :: 2.6',
            'Programming Language :: Python :: 2.7',
            'Programming Language :: Python :: 3',
            'Programming Language :: Python :: 3.3',
            'Programming Language :: Python :: 3.4',
            'Programming Language :: Python :: 3.5',
            'Programming Language :: Python :: 3.6',
        ],
        keywords='pipeline service stream pypeline',
        # unit tests
        # test_suite='test_pypelined',
        # use unittest backport to have subTest etc.
        # tests_require=['unittest2'] if sys.version_info < (3, 4) else [],
    )



from __future__ import absolute_import
import os
import sys
import logging.handlers


def configure_logging(log_level, log_format, log_dest):
    """
    Configure logging from CLI options

    :param log_level: logging verbosity
    :type log_level: int or str
    :param log_format: format string for messages
    :type log_format: str
    :param log_dest: where to send log message to
    :type log_dest: tuple[str]

    Each element in `destinations` must be either a stream name
    (`"stdout"` or `"stderr"`), or it is interpreted as a file name.
    """
    try:
        log_level = getattr(logging, log_level.upper())
    except AttributeError:
        pass
    log_level = int(log_level)
    root_handlers = logging.getLogger().handlers[:]
    logging.basicConfig(format=log_format, level=log_level)
    root_fmt = logging.getLogger().handlers[0].formatter
    # we add handlers by ourselves to use appropriate classes
    # during initialisation, use the default handler in case something goes wrong
    for destination in log_dest:
        if destination == 'stderr':
            root_handlers.append(logging.StreamHandler(sys.stderr))
        elif destination == 'stdout':
            root_handlers.append(logging.StreamHandler(sys.stdout))
        else:
            if not os.path.isdir(os.path.dirname(destination)):
                os.makedirs(os.path.dirname(destination))
            root_handlers.append(logging.handlers.WatchedFileHandler(filename=destination))
        root_handlers[-1].setFormatter(root_fmt)
    logging.getLogger().handlers[:] = root_handlers



from __future__ import absolute_import, division
import time

import chainlet


def _line_format(name, tags, fields, timestamp=None):
    """
    Format a report as InfluxDB line format

    :param name: name of the report
    :type name: str
    :param tags: tags identifying the specific report
    :type tags: dict[str]
    :param fields: measurements of the report
    :type fields: dict[str]
    :param timestamp: when the measurement was taken, in **seconds** since the epoch
    :type timestamp: float, int or None
    """
    output_str = name
    if tags:
        output_str += ',' + ','.join('%s=%s' % (key, value) for key, value in sorted(tags.items()))
    output_str += ' '
    output_str += ','.join(('%s=%r' % (key, value)).replace("'", '"') for key, value in sorted(fields.items()))
    if timestamp is not None:
        # line protocol requires nanosecond precision, python uses seconds
        output_str += ' %d' % (timestamp * 1E9)
    return output_str + '\n'


@chainlet.genlet
def telegraf_message(name, static_tags=None, dynamic_tags=(), fields=None, time_resolution=1):
    """
    Convert mapping data to line format reports suitable for telegraf

    :param name: name of the measurement
    :type name: str
    :param static_tags: predefined tags to identify the measurement, e.g. ``{'location': 'hawaii'}``
    :type static_tags: dict[str, str]
    :param dynamic_tags: keys to read from each data chunk and add as tags
    :type dynamic_tags: set[str], list[str], tuple[str]
    :param fields: keys of the report to pass on as fields; if :py:const:`None`, pass on all non-tag keys
    :type fields: set[str], list[str], tuple[str] or None
    :param time_resolution: resolution at which timestamps are reported, in seconds
    :type time_resolution: int or float
    """
    static_tags = static_tags or {}
    report = yield
    while True:
        message_tags = static_tags.copy()
        message_fields = {}
        for key in report:
            if key in dynamic_tags:
                message_tags[key] = report[key]
            elif fields is None or key in fields:
                message_fields[key] = report[key]
        message = _line_format(
                name=name % report,
                tags=message_tags,
                fields=message_fields,
                timestamp=(time.time() // time_resolution)*time_resolution
            )
        report = yield message



import chainlet


@chainlet.genlet
def remap(key_map, cull_unknown=True):
    """
    Map existing keys to new keys using a mapping

    :param key_map: mapping from old to new keys
    :type key_map: :py:class:`dict` or :py:class:`~collections.Mapping`
    :param cull_unknown: whether to remove all unmapped keys
    :type cull_unknown: bool
    :rtype: Generator[Dict, Dict, None]
    """
    input_dict = yield
    while True:
        output_dict = {}
        for key, val in input_dict.items():
            try:
                new_key = key_map[key]
            except KeyError:
                if not cull_unknown:
                    output_dict[key] = val
            else:
                output_dict[new_key] = val
        input_dict = yield output_dict


@chainlet.funclet
def update(value, iterable=None, **kwargs):
    """
    Insert ``key: value`` pairs into the report

    :param value: data chunk to operator on
    :type value: :py:class:`dict` or :py:class:`~collections.Mapping`
    :param iterable: iterable of ``(key, value)`` pairs
    :type iterable: iterable[(str, T)]
    :param kwargs: explicit ``key=value`` parameters
    """
    value = value.copy()
    if iterable:
        value.update(iterable, **kwargs)
    else:
        value.update(**kwargs)
    return value



from __future__ import absolute_import
import socket

import chainlet


class BaseSocket(chainlet.ChainLink):
    def __init__(self, host, port, encoding='utf-8'):
        self._address = (host, port)
        self.encoding = encoding

    @property
    def host(self):
        """Host messages are sent to"""
        return self._address[0]

    def port(self):
        """Port messages are sent to"""
        return self._address[1]

    def _encode(self, message):
        if self.encoding:
            return message.encode(self.encoding)
        return message

    def __repr__(self):
        if self.encoding:
            return '%s(%s, %s, %s)' % (self.__class__.__name__, self.host, self.port, self.encoding)
        return '%s(%s, %s)' % (self.__class__.__name__, self.host, self.port)


class UDPSocket(BaseSocket):
    """
    Chainable socket that sends data chunks via UDP using IPv4

    :param host: host to send messages to as a name or IPv4 address
    :type host: str
    :param port: port to send messages to
    :type port: int
    """
    def __init__(self, host, port):
        super(UDPSocket, self).__init__(host, port)
        self._socket = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)

    def chainlet_send(self, value=None):
        """Send pipeline value to ``host:port`` without consuming it"""
        message = self._encode(value)
        while message:
            message = message[self._socket.sendto(message, self._address):]
        return value


class UDP6Socket(UDPSocket):
    """
    Chainable socket that sends data chunks via UDP using IPv6

    :param host: host to send messages to as a name or IPv6 address
    :type host: str
    :param port: port to send messages to
    :type port: int
    """
    def __init__(self, host, port):
        super(UDP6Socket, self).__init__(host, port)
        self._socket = socket.socket(socket.AF_INET6, socket.SOCK_DGRAM)

udp_send = UDPSocket
udp6_send = UDP6Socket

__all__ = ['udp_send', 'udp6_send']
